from bbot.errors import HttpCompareError
from bbot.modules.base import BaseModule

"""
Port of https://github.com/iamj0ker/bypass-403/ and https://portswigger.net/bappstore/444407b96d9c4de0adb7aed89e826122
"""

# ([string]method,[string]path,[dictionary]header,[bool]strip trailing slash)
signatures = [
    ("GET", "{scheme}://{netloc}/%2e/{path}", None, False),
    ("GET", "{scheme}://{netloc}/{path}", {"X-Original-URL": "{path}"}, False),
    ("GET", "{scheme}://{netloc}/{path}", {"X-Forwarded-For": "http://127.0.0.1"}, False),
    ("GET", "{scheme}://{netloc}/{path}", {"X-rewrite-url": "nonsense"}, False),
    ("GET", "{scheme}://{netloc}/{path}.html", None, False),
    ("GET", "{scheme}://{netloc}/{path}#", None, False),
    ("POST", "{scheme}://{netloc}/{path}", {"Content-Length": "0"}, False),
    ("GET", "{scheme}://{netloc}/{path}.php", None, False),
    ("GET", "{scheme}://{netloc}/{path}.json", None, False),
    ("TRACE", "{scheme}://{netloc}/{path}", None, True),
    ("GET", "{scheme}://{netloc}/(S(X))/{path}", None, True),  # ASPNET COOKIELESS URLS
    ("GET", "{scheme}://{netloc}/(S(X))/../(S(X))/{path}", None, True),  # ASPNET COOKIELESS URLS
]


query_payloads = [
    "%09",
    "%20",
    "%23",
    "%2e",
    "%2f",
    ".",
    "?",
    ";",
    "..;",
    ";%09",
    ";%09..",
    ";%09..;",
    ";%2f..",
    "*",
    "/*",
    "..;/",
    ";/",
    "/..;/",
    "/;/",
    "/./",
    "//",
    "/.",
    "/?anything",
]

header_payloads = {
    "Client-IP": "127.0.0.1",
    "X-Real-Ip": "127.0.0.1",
    "Redirect": "127.0.0.1",
    "Referer": "127.0.0.1",
    "X-Client-IP": "127.0.0.1",
    "X-Custom-IP-Authorization": "127.0.0.1",
    "X-Forwarded-By": "127.0.0.1",
    "X-Forwarded-For": "127.0.0.1",
    "X-Forwarded-Host": "127.0.0.1",
    "X-Forwarded-Port": "80",
    "X-True-IP": "127.0.0.1",
    "X-Host": "127.0.0.1",
}

# This is planned to be replaced in the future: https://github.com/blacklanternsecurity/bbot/issues/1068
waf_strings = ["The requested URL was rejected"]

for qp in query_payloads:
    signatures.append(("GET", "{scheme}://{netloc}/{path}%s" % qp, None, True))
    if "?" not in qp:  # we only want to use "?" after the path
        signatures.append(("GET", "{scheme}://{netloc}/%s/{path}" % qp, None, True))

for hp_key in header_payloads.keys():
    signatures.append(("GET", "{scheme}://{netloc}/{path}", {hp_key: header_payloads[hp_key]}, False))


class bypass403(BaseModule):
    watched_events = ["URL"]
    produced_events = ["FINDING"]
    flags = ["active", "aggressive", "web-thorough"]
    meta = {"description": "Check 403 pages for common bypasses", "created_date": "2022-07-05", "author": "@liquidsec"}
    in_scope_only = True

    async def do_checks(self, compare_helper, event, collapse_threshold):
        results = set()
        error_count = 0

        for sig in signatures:
            if error_count > 3:
                self.warning(f"Received too many errors for URL {event.data} aborting bypass403")
                return None

            sig = self.format_signature(sig, event)
            if sig[2] != None:
                headers = dict(sig[2])
            else:
                headers = None
            try:
                match, reasons, reflection, subject_response = await compare_helper.compare(
                    sig[1], headers=headers, method=sig[0], allow_redirects=True
                )
            except HttpCompareError as e:
                error_count += 1
                self.debug(e)
                continue

            # In some cases WAFs will respond with a 200 code which causes a false positive
            if subject_response != None:
                for ws in waf_strings:
                    if ws in subject_response.text:
                        self.debug("Rejecting result based on presence of WAF string")
                        return

            if match == False:
                if str(subject_response.status_code)[0] != "4":
                    if sig[2]:
                        added_header_tuple = next(iter(sig[2].items()))
                        reported_signature = f"Added Header: {added_header_tuple[0]}: {added_header_tuple[1]}"
                    else:
                        reported_signature = f"Modified URL: {sig[0]} {sig[1]}"
                    description = f"403 Bypass Reasons: [{','.join(reasons)}] Sig: [{reported_signature}]"
                    results.add(description)
                    if len(results) > collapse_threshold:
                        return results
                else:
                    self.debug(f"Status code changed to {str(subject_response.status_code)}, ignoring")
        return results

    async def handle_event(self, event):
        try:
            compare_helper = self.helpers.http_compare(event.data, allow_redirects=True)
        except HttpCompareError as e:
            self.debug(e)
            return

        collapse_threshold = 6
        results = await self.do_checks(compare_helper, event, collapse_threshold)
        if results is None:
            return
        if len(results) > collapse_threshold:
            await self.emit_event(
                {
                    "description": f"403 Bypass MULTIPLE SIGNATURES (exceeded threshold {str(collapse_threshold)})",
                    "host": str(event.host),
                    "url": event.data,
                },
                "FINDING",
                parent=event,
                context=f"{{module}} discovered multiple potential 403 bypasses ({{event.type}}) for {event.data}",
            )
        else:
            for description in results:
                await self.emit_event(
                    {"description": description, "host": str(event.host), "url": event.data},
                    "FINDING",
                    parent=event,
                    context=f"{{module}} discovered potential 403 bypass ({{event.type}}) for {event.data}",
                )

    # When a WAF-check helper is available in the future, we will convert to HTTP_RESPONSE and check for the WAF string here.
    async def filter_event(self, event):
        if ("status-403" in event.tags) or ("status-401" in event.tags):
            return True
        return False

    def format_signature(self, sig, event):
        if sig[3] == True:
            cleaned_path = event.parsed_url.path.strip("/")
        else:
            cleaned_path = event.parsed_url.path.lstrip("/")
        kwargs = {"scheme": event.parsed_url.scheme, "netloc": event.parsed_url.netloc, "path": cleaned_path}
        formatted_url = sig[1].format(**kwargs)
        if sig[2] != None:
            formatted_headers = {k: v.format(**kwargs) for k, v in sig[2].items()}
        else:
            formatted_headers = None
        return (sig[0], formatted_url, formatted_headers)



from ..bbot_fixtures import *  # noqa: F401
from ..test_step_2.module_tests.base import ModuleTestBase


class TestScopeBaseline(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx"]

    async def setup_after_prep(self, module_test):
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        assert len(events) == 5
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "URL_UNVERIFIED"
                and str(e.host) == "127.0.0.1"
                and e.scope_distance == 0
                and "target" in e.tags
            ]
        )
        # we have two of these because the host module considers "always_emit" in its outgoing deduplication
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "IP_ADDRESS"
                and e.data == "127.0.0.1"
                and e.scope_distance == 0
                and str(e.module) == "host"
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "HTTP_RESPONSE"
                and str(e.host) == "127.0.0.1"
                and e.port == 8888
                and e.scope_distance == 0
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "URL" and str(e.host) == "127.0.0.1" and e.port == 8888 and e.scope_distance == 0
            ]
        )


class TestScopeBlacklist(TestScopeBaseline):
    blacklist = ["127.0.0.1"]

    async def setup_after_prep(self, module_test):
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        assert len(events) == 1
        assert not any(e.type == "URL" for e in events)
        assert not any(str(e.host) == "127.0.0.1" for e in events)


class TestScopeWhitelist(TestScopeBlacklist):
    blacklist = []
    whitelist = ["255.255.255.255"]

    def check(self, module_test, events):
        assert len(events) == 3
        assert not any(e.type == "URL" for e in events)
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "IP_ADDRESS" and e.data == "127.0.0.1" and e.scope_distance == 1 and "target" in e.tags
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "URL_UNVERIFIED"
                and str(e.host) == "127.0.0.1"
                and e.scope_distance == 1
                and "target" in e.tags
            ]
        )



import json
from bbot.modules.base import BaseModule


class wpscan(BaseModule):
    watched_events = ["HTTP_RESPONSE", "TECHNOLOGY"]
    produced_events = ["URL_UNVERIFIED", "FINDING", "VULNERABILITY", "TECHNOLOGY"]
    flags = ["active", "aggressive"]
    meta = {
        "description": "Wordpress security scanner. Highly recommended to use an API key for better results.",
        "created_date": "2024-05-29",
        "author": "@domwhewell-sage",
    }

    options = {
        "api_key": "",
        "enumerate": "vp,vt,tt,cb,dbe,u,m",
        "threads": 5,
        "request_timeout": 60,
        "connection_timeout": 30,
        "disable_tls_checks": True,
        "force": False,
    }
    options_desc = {
        "api_key": "WPScan API Key",
        "enumerate": "Enumeration Process see wpscan help documentation (default: vp,vt,tt,cb,dbe,u,m)",
        "threads": "How many wpscan threads to spawn (default is 5)",
        "request_timeout": "The request timeout in seconds (default 60)",
        "connection_timeout": "The connection timeout in seconds (default 30)",
        "disable_tls_checks": "Disables the SSL/TLS certificate verification (Default True)",
        "force": "Do not check if the target is running WordPress or returns a 403",
    }
    deps_apt = ["curl", "make", "gcc"]
    deps_ansible = [
        {
            "name": "Install Ruby Deps (Debian/Ubuntu)",
            "package": {"name": ["ruby-rubygems", "ruby-dev"], "state": "present"},
            "become": True,
            "when": "ansible_facts['os_family'] == 'Debian'",
        },
        {
            "name": "Install Ruby Deps (Arch)",
            "package": {"name": ["rubygems"], "state": "present"},
            "become": True,
            "when": "ansible_facts['os_family'] == 'Archlinux'",
        },
        {
            "name": "Install Ruby Deps (Fedora)",
            "package": {"name": ["rubygems", "ruby-devel"], "state": "present"},
            "become": True,
            "when": "ansible_facts['os_family'] == 'Fedora'",
        },
        {
            "name": "Install wpscan gem",
            "gem": {"name": "wpscan", "state": "latest", "user_install": False},
            "become": True,
        },
    ]

    async def setup(self):
        self.processed = set()
        self.ignore_events = ["xmlrpc", "readme"]
        self.api_key = self.config.get("api_key", "")
        self.enumerate = self.config.get("enumerate", "vp,vt,tt,cb,dbe,u,m")
        self.proxy = self.scan.config.get("http_proxy", "")
        self.threads = self.config.get("threads", 5)
        self.request_timeout = self.config.get("request_timeout", 60)
        self.connection_timeout = self.config.get("connection_timeout", 30)
        self.disable_tls_checks = self.config.get("disable_tls_checks", True)
        self.force = self.config.get("force", False)
        return True

    async def filter_event(self, event):
        host_hash = hash(event.host)
        if host_hash in self.processed:
            return False, "Host has already been processed"
        if event.type == "HTTP_RESPONSE":
            is_redirect = str(event.data["status_code"]).startswith("30")
            if is_redirect:
                return False, "URL is a redirect"
        elif event.type == "TECHNOLOGY":
            if not event.data["technology"].lower().startswith("wordpress"):
                return False, "technology is not wordpress"
        self.processed.add(host_hash)
        return True

    async def handle_event(self, event):
        if event.type == "HTTP_RESPONSE":
            await self.handle_http_response(event)
        elif event.type == "TECHNOLOGY":
            await self.handle_technology(event)

    async def handle_http_response(self, source_event):
        url = source_event.parsed_url._replace(path="/").geturl()
        command = self.construct_command(url)
        output = await self.run_process(command)
        for new_event in self.parse_wpscan_output(output.stdout, url, source_event):
            await self.emit_event(new_event)

    async def handle_technology(self, source_event):
        url = self.get_base_url(source_event)
        command = self.construct_command(url)
        output = await self.run_process(command)
        for new_event in self.parse_wpscan_output(output.stdout, url, source_event):
            await self.emit_event(new_event)

    def construct_command(self, url):
        # base executable
        command = ["wpscan", "--url", url]
        # proxy
        if self.proxy:
            command += ["--proxy", str(self.proxy)]
        # user agent
        command += ["--user-agent", f"'{self.scan.useragent}'"]
        # threads
        command += ["--max-threads", str(self.threads)]
        # request timeout
        command += ["--request-timeout", str(self.request_timeout)]
        # connection timeout
        command += ["--connect-timeout", str(self.connection_timeout)]
        # api key
        if self.api_key:
            command += ["--api-token", f"{self.api_key}"]
        # enumerate
        command += ["--enumerate", self.enumerate]
        # disable tls checks
        if self.disable_tls_checks:
            command += ["--disable-tls-checks"]
        # force
        if self.force:
            command += ["--force"]
        # output format
        command += ["--format", "json"]
        return command

    def parse_wpscan_output(self, output, base_url, source_event):
        json_output = json.loads(output)
        interesting_json = json_output.get("interesting_findings", {})
        version_json = json_output.get("version", {})
        theme_json = json_output.get("main_theme", {})
        plugins_json = json_output.get("plugins", {})
        if interesting_json:
            yield from self.parse_wp_misc(interesting_json, base_url, source_event)
        if version_json:
            yield from self.parse_wp_version(version_json, base_url, source_event)
        if theme_json:
            yield from self.parse_wp_themes(theme_json, base_url, source_event)
        if plugins_json:
            yield from self.parse_wp_plugins(plugins_json, base_url, source_event)

    def parse_wp_misc(self, interesting_json, base_url, source_event):
        for finding in interesting_json:
            url = finding.get("url", base_url)
            type = finding["type"]
            if type in self.ignore_events:
                continue
            description_string = finding["to_s"]
            interesting_entries = finding["interesting_entries"]
            if type == "headers":
                for header in interesting_entries:
                    yield self.make_event(
                        {"technology": str(header).lower(), "url": url, "host": str(source_event.host)},
                        "TECHNOLOGY",
                        source_event,
                    )
            else:
                url_event = self.make_event(url, "URL_UNVERIFIED", parent=source_event, tags=["httpx-safe"])
                if url_event:
                    yield url_event
                yield self.make_event(
                    {"description": description_string, "url": url, "host": str(source_event.host)},
                    "FINDING",
                    source_event,
                )

    def parse_wp_version(self, version_json, url, source_event):
        version = version_json.get("number", "")
        if version:
            technology = f"wordpress {version}"
        else:
            technology = "wordpress detect"
        yield self.make_event(
            {"technology": str(technology).lower(), "url": url, "host": str(source_event.host)},
            "TECHNOLOGY",
            source_event,
        )
        for wp_vuln in version_json.get("vulnerabilities", []):
            yield self.make_event(
                {
                    "severity": "HIGH",
                    "host": str(source_event.host),
                    "url": url,
                    "description": self.vulnerability_to_s(wp_vuln),
                },
                "VULNERABILITY",
                source_event,
            )

    def parse_wp_themes(self, theme_json, url, source_event):
        name = theme_json.get("slug", "")
        version = theme_json.get("version", {}).get("number", "")
        if name:
            if version:
                technology = f"{name} v{version}"
            else:
                technology = name
        yield self.make_event(
            {"technology": str(technology).lower(), "url": url, "host": str(source_event.host)},
            "TECHNOLOGY",
            source_event,
        )
        for theme_vuln in theme_json.get("vulnerabilities", []):
            yield self.make_event(
                {
                    "severity": "HIGH",
                    "host": str(source_event.host),
                    "url": url,
                    "description": self.vulnerability_to_s(theme_vuln),
                },
                "VULNERABILITY",
                source_event,
            )

    def parse_wp_plugins(self, plugins_json, base_url, source_event):
        for name, plugin in plugins_json.items():
            url = plugin.get("location", base_url)
            if url != base_url:
                url_event = self.make_event(url, "URL_UNVERIFIED", parent=source_event, tags=["httpx-safe"])
                if url_event:
                    yield url_event
            version = plugin.get("version", {}).get("number", "")
            if version:
                technology = f"{name} {version}"
            else:
                technology = name
            yield self.make_event(
                {"technology": str(technology).lower(), "url": url, "host": str(source_event.host)},
                "TECHNOLOGY",
                source_event,
            )
            for vuln in plugin.get("vulnerabilities", []):
                yield self.make_event(
                    {
                        "severity": "HIGH",
                        "host": str(source_event.host),
                        "url": url,
                        "description": self.vulnerability_to_s(vuln),
                    },
                    "VULNERABILITY",
                    source_event,
                )

    def vulnerability_to_s(self, vuln_json):
        string = []
        title = vuln_json.get("title", "")
        string.append(f"Title: {title}")
        fixed_in = vuln_json.get("fixed_in", "")
        string.append(f"Fixed in: {fixed_in}")
        references = vuln_json.get("references", {})
        if references:
            cves = references.get("cve", [])
            urls = references.get("url", [])
            youtube_urls = references.get("youtube", [])
            cves_list = []
            for cve in cves:
                cves_list.append(f"CVE-{cve}")
            if cves_list:
                string.append(f"CVEs: [{', '.join(cves_list)}]")
            if urls:
                string.append(f"References: [{', '.join(urls)}]")
            if youtube_urls:
                string.append(f"Youtube Links: [{', '.join(youtube_urls)}]")
        return " ".join(string)

    def get_base_url(self, event):
        base_url = event.data.get("url", "")
        if not base_url:
            base_url = f"https://{event.host}"
        return self.helpers.urlparse(base_url)._replace(path="/").geturl()



import httpx

from .base import ModuleTestBase


class TestDiscord(ModuleTestBase):
    targets = ["http://127.0.0.1:8888/cookie.aspx", "http://127.0.0.1:8888/cookie2.aspx", "foo.bar"]
    modules_overrides = ["discord", "excavate", "badsecrets", "httpx"]

    webhook_url = "https://discord.com/api/webhooks/1234/deadbeef-P-uF-asdf"
    config_overrides = {"modules": {"discord": {"webhook_url": webhook_url}}}

    def custom_setup(self, module_test):
        respond_args = {
            "response_data": '<html><body><p>Express Cookie Test<a href="ftp://asdf.foo.bar/asdf.txt"/></p></body></html>',
            "headers": {
                "set-cookie": "connect.sid=s%3A8FnPwdeM9kdGTZlWvdaVtQ0S1BCOhY5G.qys7H2oGSLLdRsEq7sqh7btOohHsaRKqyjV4LiVnBvc; Path=/; Expires=Wed, 05 Apr 2023 04:47:29 GMT; HttpOnly"
            },
        }
        module_test.set_expect_requests(expect_args={"uri": "/cookie.aspx"}, respond_args=respond_args)
        module_test.set_expect_requests(expect_args={"uri": "/cookie2.aspx"}, respond_args=respond_args)
        module_test.request_count = 0

    async def setup_after_prep(self, module_test):
        self.custom_setup(module_test)

        def custom_response(request: httpx.Request):
            module_test.request_count += 1
            if module_test.request_count == 2:
                return httpx.Response(status_code=429, json={"retry_after": 0.01})
            else:
                return httpx.Response(status_code=module_test.module.good_status_code)

        module_test.httpx_mock.add_callback(custom_response, url=self.webhook_url)

    def check(self, module_test, events):
        vulns = [e for e in events if e.type == "VULNERABILITY"]
        findings = [e for e in events if e.type == "FINDING"]
        assert len(findings) == 1
        assert len(vulns) == 2
        assert module_test.request_count == 4



from .base import ModuleTestBase


class TestDnsbrute_mutations(ModuleTestBase):
    targets = [
        "blacklanternsecurity.com",
        "rrrr.blacklanternsecurity.com",
        "asdff-ffdsa.blacklanternsecurity.com",
        "hmmmm.test1.blacklanternsecurity.com",
        "uuuuu.test2.blacklanternsecurity.com",
    ]

    async def setup_after_prep(self, module_test):

        old_run_live = module_test.scan.helpers.run_live

        async def new_run_live(*command, check=False, text=True, **kwargs):
            if "massdns" in command[:2]:
                _input = [l async for l in kwargs["input"]]
                if "rrrr-test.blacklanternsecurity.com" in _input:
                    yield """{"name": "rrrr-test.blacklanternsecurity.com.", "type": "A", "class": "IN", "status": "NOERROR", "rx_ts": 1713974911725326170, "data": {"answers": [{"ttl": 86400, "type": "A", "class": "IN", "name": "rrrr-test.blacklanternsecurity.com.", "data": "1.2.3.4."}]}, "flags": ["rd", "ra"], "resolver": "195.226.187.130:53", "proto": "UDP"}"""
                if "rrrr-ffdsa.blacklanternsecurity.com" in _input:
                    yield """{"name": "rrrr-ffdsa.blacklanternsecurity.com.", "type": "A", "class": "IN", "status": "NOERROR", "rx_ts": 1713974911725326170, "data": {"answers": [{"ttl": 86400, "type": "A", "class": "IN", "name": "rrrr-ffdsa.blacklanternsecurity.com.", "data": "1.2.3.4."}]}, "flags": ["rd", "ra"], "resolver": "195.226.187.130:53", "proto": "UDP"}"""
                if "hmmmm.test2.blacklanternsecurity.com" in _input:
                    yield """{"name": "hmmmm.test2.blacklanternsecurity.com.", "type": "A", "class": "IN", "status": "NOERROR", "rx_ts": 1713974911725326170, "data": {"answers": [{"ttl": 86400, "type": "A", "class": "IN", "name": "hmmmm.test2.blacklanternsecurity.com.", "data": "1.2.3.4."}]}, "flags": ["rd", "ra"], "resolver": "195.226.187.130:53", "proto": "UDP"}"""
            else:
                async for _ in old_run_live(*command, check=False, text=True, **kwargs):
                    yield _

        module_test.monkeypatch.setattr(module_test.scan.helpers, "run_live", new_run_live)

        await module_test.mock_dns(
            {
                "blacklanternsecurity.com": {"A": ["1.2.3.4"]},
                # targets
                "rrrr.blacklanternsecurity.com": {"A": ["1.2.3.4"]},
                "asdff-ffdsa.blacklanternsecurity.com": {"A": ["1.2.3.4"]},
                "hmmmm.test1.blacklanternsecurity.com": {"A": ["1.2.3.4"]},
                "uuuuu.test2.blacklanternsecurity.com": {"A": ["1.2.3.4"]},
                # devops mutation
                "rrrr-test.blacklanternsecurity.com": {"A": ["1.2.3.4"]},
                # target-specific dns mutation
                "rrrr-ffdsa.blacklanternsecurity.com": {"A": ["1.2.3.4"]},
                # subdomain from one subdomain on a different subdomain
                "hmmmm.test2.blacklanternsecurity.com": {"A": ["1.2.3.4"]},
            }
        )

    def check(self, module_test, events):
        assert len(events) == 9
        assert 1 == len(
            [
                e
                for e in events
                if e.data == "rrrr-test.blacklanternsecurity.com" and str(e.module) == "dnsbrute_mutations"
            ]
        ), "Failed to find devops mutation (word_cloud)"
        assert 1 == len(
            [
                e
                for e in events
                if e.data == "rrrr-ffdsa.blacklanternsecurity.com" and str(e.module) == "dnsbrute_mutations"
            ]
        ), "Failed to find target-specific mutation (word_cloud.dns_mutator)"
        assert 1 == len(
            [
                e
                for e in events
                if e.data == "hmmmm.test2.blacklanternsecurity.com" and str(e.module) == "dnsbrute_mutations"
            ]
        ), "Failed to find subdomain taken from another subdomain"



import ssl

ssl_context_noverify = ssl.create_default_context()
ssl_context_noverify.check_hostname = False
ssl_context_noverify.verify_mode = ssl.CERT_NONE
ssl_context_noverify.options &= ~ssl.OP_NO_SSLv2 & ~ssl.OP_NO_SSLv3
ssl_context_noverify.set_ciphers("ALL:@SECLEVEL=0")
ssl_context_noverify.options |= 0x4  # Add the OP_LEGACY_SERVER_CONNECT option



import dns
import logging
import dns.exception
import dns.asyncresolver
from cachetools import LFUCache
from radixtarget import RadixTarget

from bbot.errors import DNSError
from bbot.core.engine import EngineClient
from bbot.core.helpers.async_helpers import async_cachedmethod
from ..misc import clean_dns_record, is_ip, is_domain, is_dns_name

from .engine import DNSEngine

log = logging.getLogger("bbot.core.helpers.dns")


class DNSHelper(EngineClient):

    SERVER_CLASS = DNSEngine
    ERROR_CLASS = DNSError

    """Helper class for DNS-related operations within BBOT.

    This class provides mechanisms for host resolution, wildcard domain detection, event tagging, and more.
    It centralizes all DNS-related activities in BBOT, offering both synchronous and asynchronous methods
    for DNS resolution, as well as various utilities for batch resolution and DNS query filtering.

    Attributes:
        parent_helper: A reference to the instantiated `ConfigAwareHelper` (typically `scan.helpers`).
        resolver (BBOTAsyncResolver): An asynchronous DNS resolver tailored for BBOT with rate-limiting capabilities.
        timeout (int): The timeout value for DNS queries. Defaults to 5 seconds.
        retries (int): The number of retries for failed DNS queries. Defaults to 1.
        abort_threshold (int): The threshold for aborting after consecutive failed queries. Defaults to 50.
        runaway_limit (int): Maximum allowed distance for consecutive DNS resolutions. Defaults to 5.
        all_rdtypes (list): A list of DNS record types to be considered during operations.
        wildcard_ignore (tuple): Domains to be ignored during wildcard detection.
        wildcard_tests (int): Number of tests to be run for wildcard detection. Defaults to 5.
        _wildcard_cache (dict): Cache for wildcard detection results.
        _dns_cache (LRUCache): Cache for DNS resolution results, limited in size.
        resolver_file (Path): File containing system's current resolver nameservers.
        filter_bad_ptrs (bool): Whether to filter out DNS names that appear to be auto-generated PTR records. Defaults to True.

    Args:
        parent_helper: The parent helper object with configuration details and utilities.

    Raises:
        DNSError: If an issue arises when creating the BBOTAsyncResolver instance.

    Examples:
        >>> dns_helper = DNSHelper(parent_config)
        >>> resolved_host = dns_helper.resolver.resolve("example.com")
    """

    def __init__(self, parent_helper):
        self.parent_helper = parent_helper
        self.config = self.parent_helper.config
        self.dns_config = self.config.get("dns", {})
        engine_debug = self.config.get("engine", {}).get("debug", False)
        super().__init__(server_kwargs={"config": self.config}, debug=engine_debug)

        # resolver
        self.timeout = self.dns_config.get("timeout", 5)
        self.resolver = dns.asyncresolver.Resolver()
        self.resolver.rotate = True
        self.resolver.timeout = self.timeout
        self.resolver.lifetime = self.timeout

        self.runaway_limit = self.config.get("runaway_limit", 5)

        # wildcard handling
        self.wildcard_disable = self.dns_config.get("wildcard_disable", False)
        self.wildcard_ignore = RadixTarget()
        for d in self.dns_config.get("wildcard_ignore", []):
            self.wildcard_ignore.insert(d)

        # copy the system's current resolvers to a text file for tool use
        self.system_resolvers = dns.resolver.Resolver().nameservers
        # TODO: DNS server speed test (start in background task)
        self.resolver_file = self.parent_helper.tempfile(self.system_resolvers, pipe=False)

        # brute force helper
        self._brute = None

        self._is_wildcard_cache = LFUCache(maxsize=1000)
        self._is_wildcard_domain_cache = LFUCache(maxsize=1000)

    async def resolve(self, query, **kwargs):
        return await self.run_and_return("resolve", query=query, **kwargs)

    async def resolve_raw(self, query, **kwargs):
        return await self.run_and_return("resolve_raw", query=query, **kwargs)

    async def resolve_batch(self, queries, **kwargs):
        agen = self.run_and_yield("resolve_batch", queries=queries, **kwargs)
        while 1:
            try:
                yield await agen.__anext__()
            except (StopAsyncIteration, GeneratorExit):
                await agen.aclose()
                break

    async def resolve_raw_batch(self, queries):
        agen = self.run_and_yield("resolve_raw_batch", queries=queries)
        while 1:
            try:
                yield await agen.__anext__()
            except (StopAsyncIteration, GeneratorExit):
                await agen.aclose()
                break

    @property
    def brute(self):
        if self._brute is None:
            from .brute import DNSBrute

            self._brute = DNSBrute(self.parent_helper)
        return self._brute

    @async_cachedmethod(lambda self: self._is_wildcard_cache)
    async def is_wildcard(self, query, ips=None, rdtype=None):
        """
        Use this method to check whether a *host* is a wildcard entry

        This can reliably tell the difference between a valid DNS record and a wildcard within a wildcard domain.

        If you want to know whether a domain is using wildcard DNS, use `is_wildcard_domain()` instead.

        Args:
            query (str): The hostname to check for a wildcard entry.
            ips (list, optional): List of IPs to compare against, typically obtained from a previous DNS resolution of the query.
            rdtype (str, optional): The DNS record type (e.g., "A", "AAAA") to consider during the check.

        Returns:
            dict: A dictionary indicating if the query is a wildcard for each checked DNS record type.
                Keys are DNS record types like "A", "AAAA", etc.
                Values are tuples where the first element is a boolean indicating if the query is a wildcard,
                and the second element is the wildcard parent if it's a wildcard.

        Raises:
            ValueError: If only one of `ips` or `rdtype` is specified or if no valid IPs are specified.

        Examples:
            >>> is_wildcard("www.github.io")
            {"A": (True, "github.io"), "AAAA": (True, "github.io")}

            >>> is_wildcard("www.evilcorp.com", ips=["93.184.216.34"], rdtype="A")
            {"A": (False, "evilcorp.com")}

        Note:
            `is_wildcard` can be True, False, or None (indicating that wildcard detection was inconclusive)
        """
        if [ips, rdtype].count(None) == 1:
            raise ValueError("Both ips and rdtype must be specified")

        query = self._wildcard_prevalidation(query)
        if not query:
            return {}

        # skip check if the query is a domain
        if is_domain(query):
            return {}

        return await self.run_and_return("is_wildcard", query=query, ips=ips, rdtype=rdtype)

    @async_cachedmethod(lambda self: self._is_wildcard_domain_cache)
    async def is_wildcard_domain(self, domain, log_info=False):
        domain = self._wildcard_prevalidation(domain)
        if not domain:
            return {}

        return await self.run_and_return("is_wildcard_domain", domain=domain, log_info=False)

    def _wildcard_prevalidation(self, host):
        if self.wildcard_disable:
            return False

        host = clean_dns_record(host)
        # skip check if it's an IP or a plain hostname
        if is_ip(host) or not "." in host:
            return False

        # skip if query isn't a dns name
        if not is_dns_name(host):
            return False

        # skip check if the query's parent domain is excluded in the config
        wildcard_ignore = self.wildcard_ignore.search(host)
        if wildcard_ignore:
            log.debug(f"Skipping wildcard detection on {host} because {wildcard_ignore} is excluded in the config")
            return False

        return host

    async def _mock_dns(self, mock_data):
        from .mock import MockResolver

        self.resolver = MockResolver(mock_data)
        await self.run_and_return("_mock_dns", mock_data=mock_data)



from pathlib import Path
from .base import ModuleTestBase


class TestUnstructured(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["unstructured", "filedownload", "httpx", "excavate", "speculate"]
    config_overrides = {"web": {"spider_distance": 2, "spider_depth": 2}}

    pdf_data = r"""%PDF-1.3
%���� ReportLab Generated PDF document http://www.reportlab.com
1 0 obj
<<
/F1 2 0 R
>>
endobj
2 0 obj
<<
/BaseFont /Helvetica /Encoding /WinAnsiEncoding /Name /F1 /Subtype /Type1 /Type /Font
>>
endobj
3 0 obj
<<
/Contents 7 0 R /MediaBox [ 0 0 595.2756 841.8898 ] /Parent 6 0 R /Resources <<
/Font 1 0 R /ProcSet [ /PDF /Text /ImageB /ImageC /ImageI ]
>> /Rotate 0 /Trans <<

>> 
  /Type /Page
>>
endobj
4 0 obj
<<
/PageMode /UseNone /Pages 6 0 R /Type /Catalog
>>
endobj
5 0 obj
<<
/Author (anonymous) /CreationDate (D:20240603185816+00'00') /Creator (ReportLab PDF Library - www.reportlab.com) /Keywords () /ModDate (D:20240603185816+00'00') /Producer (ReportLab PDF Library - www.reportlab.com) 
  /Subject (unspecified) /Title (untitled) /Trapped /False
>>
endobj
6 0 obj
<<
/Count 1 /Kids [ 3 0 R ] /Type /Pages
>>
endobj
7 0 obj
<<
/Filter [ /ASCII85Decode /FlateDecode ] /Length 107
>>
stream
GapQh0E=F,0U\H3T\pNYT^QKk?tc>IP,;W#U1^23ihPEM_?CW4KISi90MjG^2,FS#<RC5+c,n)Z;$bK$b"5I[<!^TD#gi]&=5X,[5@Y@V~>endstream
endobj
xref
0 8
0000000000 65535 f 
0000000073 00000 n 
0000000104 00000 n 
0000000211 00000 n 
0000000414 00000 n 
0000000482 00000 n 
0000000778 00000 n 
0000000837 00000 n 
trailer
<<
/ID 
[<80d9f5b964fc99284501deb7a6a637f7><80d9f5b964fc99284501deb7a6a637f7>]
% ReportLab generated PDF document -- digest (http://www.reportlab.com)

/Info 5 0 R
/Root 4 0 R
/Size 8
>>
startxref
1034
%%EOF"""

    unstructured_response = "Hello, World!"

    async def setup_after_prep(self, module_test):
        module_test.set_expect_requests(
            dict(uri="/"),
            dict(response_data='<a href="/Test_PDF"/>'),
        )
        module_test.set_expect_requests(
            dict(uri="/Test_PDF"),
            dict(response_data=self.pdf_data, headers={"Content-Type": "application/pdf"}),
        )

    def check(self, module_test, events):
        filesystem_events = [e for e in events if e.type == "FILESYSTEM"]
        assert 1 == len(filesystem_events), filesystem_events
        filesystem_event = filesystem_events[0]
        file = Path(filesystem_event.data["path"])
        assert file.is_file(), "Destination file doesn't exist"
        assert open(file).read() == self.pdf_data, f"File at {file} does not contain the correct content"
        raw_text_events = [e for e in events if e.type == "RAW_TEXT"]
        assert 1 == len(raw_text_events), "Failed to emmit RAW_TEXT event"
        assert (
            raw_text_events[0].data == self.unstructured_response
        ), f"Text extracted from PDF is incorrect, got {raw_text_events[0].data}"



from bbot.modules.templates.subdomain_enum import subdomain_enum


class columbus(subdomain_enum):
    flags = ["subdomain-enum", "passive", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {
        "description": "Query the Columbus Project API for subdomains",
        "created_date": "2023-06-01",
        "author": "@TheTechromancer",
    }

    base_url = "https://columbus.elmasy.com/api/lookup"

    async def request_url(self, query):
        url = f"{self.base_url}/{self.helpers.quote(query)}?days=365"
        return await self.request_with_fail_count(url)

    def parse_results(self, r, query):
        results = set()
        json = r.json()
        if json and isinstance(json, list):
            return set([f"{s.lower()}.{query}" for s in json])
        return results



from ..bbot_fixtures import *


@pytest.mark.asyncio
async def test_python_api():
    from bbot import Scanner

    # make sure events are properly yielded
    scan1 = Scanner("127.0.0.1")
    events1 = []
    async for event in scan1.async_start():
        events1.append(event)
    assert any("127.0.0.1" == e for e in events1)
    # make sure output files work
    scan2 = Scanner("127.0.0.1", output_modules=["json"], scan_name="python_api_test")
    await scan2.async_start_without_generator()
    scan_home = scan2.helpers.scans_dir / "python_api_test"
    out_file = scan_home / "output.json"
    assert list(scan2.helpers.read_file(out_file))
    scan_log = scan_home / "scan.log"
    debug_log = scan_home / "debug.log"
    assert scan_log.is_file()
    assert "python_api_test" in open(scan_log).read()
    assert debug_log.is_file()
    assert "python_api_test" in open(debug_log).read()

    scan3 = Scanner("127.0.0.1", output_modules=["json"], scan_name="scan_logging_test")
    await scan3.async_start_without_generator()

    assert "scan_logging_test" not in open(scan_log).read()
    assert "scan_logging_test" not in open(debug_log).read()

    scan_home = scan3.helpers.scans_dir / "scan_logging_test"
    out_file = scan_home / "output.json"
    assert list(scan3.helpers.read_file(out_file))
    scan_log = scan_home / "scan.log"
    debug_log = scan_home / "debug.log"
    assert scan_log.is_file()
    assert debug_log.is_file()
    assert "scan_logging_test" in open(scan_log).read()
    assert "scan_logging_test" in open(debug_log).read()

    # make sure config loads properly
    bbot_home = "/tmp/.bbot_python_api_test"
    Scanner("127.0.0.1", config={"home": bbot_home})
    assert os.environ["BBOT_TOOLS"] == str(Path(bbot_home) / "tools")

    # output modules override
    scan4 = Scanner()
    assert set(scan4.preset.output_modules) == {"csv", "json", "python", "txt"}
    scan5 = Scanner(output_modules=["json"])
    assert set(scan5.preset.output_modules) == {"json"}

    # custom target types
    custom_target_scan = Scanner("ORG:evilcorp")
    events = [e async for e in custom_target_scan.async_start()]
    assert 1 == len([e for e in events if e.type == "ORG_STUB" and e.data == "evilcorp" and "target" in e.tags])

    # presets
    scan6 = Scanner("evilcorp.com", presets=["subdomain-enum"])
    assert "sslcert" in scan6.preset.modules


def test_python_api_sync():
    from bbot.scanner import Scanner

    # make sure events are properly yielded
    scan1 = Scanner("127.0.0.1")
    events1 = []
    for event in scan1.start():
        events1.append(event)
    assert any("127.0.0.1" == e for e in events1)
    # make sure output files work
    scan2 = Scanner("127.0.0.1", output_modules=["json"], scan_name="python_api_test")
    scan2.start_without_generator()
    out_file = scan2.helpers.scans_dir / "python_api_test" / "output.json"
    assert list(scan2.helpers.read_file(out_file))
    # make sure config loads properly
    bbot_home = "/tmp/.bbot_python_api_test"
    Scanner("127.0.0.1", config={"home": bbot_home})
    assert os.environ["BBOT_TOOLS"] == str(Path(bbot_home) / "tools")


def test_python_api_validation():
    from bbot.scanner import Scanner, Preset

    # invalid module
    with pytest.raises(ValidationError) as error:
        Scanner(modules=["asdf"])
    assert str(error.value) == 'Could not find scan module "asdf". Did you mean "asn"?'
    # invalid output module
    with pytest.raises(ValidationError) as error:
        Scanner(output_modules=["asdf"])
    assert str(error.value) == 'Could not find output module "asdf". Did you mean "teams"?'
    # invalid excluded module
    with pytest.raises(ValidationError) as error:
        Scanner(exclude_modules=["asdf"])
    assert str(error.value) == 'Could not find module "asdf". Did you mean "asn"?'
    # invalid flag
    with pytest.raises(ValidationError) as error:
        Scanner(flags=["asdf"])
    assert str(error.value) == 'Could not find flag "asdf". Did you mean "safe"?'
    # invalid required flag
    with pytest.raises(ValidationError) as error:
        Scanner(require_flags=["asdf"])
    assert str(error.value) == 'Could not find flag "asdf". Did you mean "safe"?'
    # invalid excluded flag
    with pytest.raises(ValidationError) as error:
        Scanner(exclude_flags=["asdf"])
    assert str(error.value) == 'Could not find flag "asdf". Did you mean "safe"?'
    # output module as normal module
    with pytest.raises(ValidationError) as error:
        Scanner(modules=["json"])
    assert str(error.value) == 'Could not find scan module "json". Did you mean "asn"?'
    # normal module as output module
    with pytest.raises(ValidationError) as error:
        Scanner(output_modules=["robots"])
    assert str(error.value) == 'Could not find output module "robots". Did you mean "web_report"?'
    # invalid preset type
    with pytest.raises(ValidationError) as error:
        Scanner(preset="asdf")
    assert str(error.value) == 'Preset must be of type Preset, not "str"'
    # include nonexistent preset
    with pytest.raises(ValidationError) as error:
        Preset(include=["asdf"])
    assert (
        str(error.value) == 'Could not find preset at "asdf" - file does not exist. Use -lp to list available presets'
    )



from .base import ModuleTestBase


class TestAsset_Inventory(ModuleTestBase):
    targets = ["127.0.0.1", "bbottest.notreal"]
    scan_name = "asset_inventory_test"
    config_overrides = {"dns": {"minimal": False}, "modules": {"portscan": {"ports": "9999"}}}
    modules_overrides = ["asset_inventory", "portscan", "sslcert"]

    masscan_output = """{   "ip": "127.0.0.1",   "timestamp": "1680197558", "ports": [ {"port": 9999, "proto": "tcp", "status": "open", "reason": "syn-ack", "ttl": 54} ] }"""

    async def setup_before_prep(self, module_test):

        async def run_masscan(command, *args, **kwargs):
            if "masscan" in command[:2]:
                targets = open(command[11]).read().splitlines()
                yield "["
                for l in self.masscan_output.splitlines():
                    if "127.0.0.1/32" in targets:
                        yield self.masscan_output
                yield "]"
            else:
                async for l in module_test.scan.helpers.run_live(command, *args, **kwargs):
                    yield l

        module_test.monkeypatch.setattr(module_test.scan.helpers, "run_live", run_masscan)

        await module_test.mock_dns(
            {
                "1.0.0.127.in-addr.arpa": {"PTR": ["www.bbottest.notreal"]},
                "www.bbottest.notreal": {"A": ["127.0.0.1"]},
            }
        )

    def check(self, module_test, events):
        assert any(e.data == "127.0.0.1:9999" for e in events), "No open port found"
        assert any(e.data == "www.bbottest.notreal" for e in events), "No DNS name found"
        filename = next(module_test.scan.home.glob("asset-inventory.csv"))
        with open(filename) as f:
            content = f.read()
            assert "www.bbottest.notreal,,,127.0.0.1" in content
        filename = next(module_test.scan.home.glob("asset-inventory-ip-addresses-table*.txt"))
        with open(filename) as f:
            assert "127.0.0.0/16" in f.read()
        filename = next(module_test.scan.home.glob("asset-inventory-domains-table*.txt"))
        with open(filename) as f:
            content = f.read()
            assert "bbottest.notreal" in content


class TestAsset_InventoryEmitPrevious(TestAsset_Inventory):
    config_overrides = {"dns": {"minimal": False}, "modules": {"asset_inventory": {"use_previous": True}}}
    modules_overrides = ["asset_inventory"]

    def check(self, module_test, events):
        assert any(e.data == "www.bbottest.notreal:9999" for e in events), "No open port found"
        assert any(e.data == "www.bbottest.notreal" for e in events), "No DNS name found"
        filename = next(module_test.scan.home.glob("asset-inventory.csv"))
        with open(filename) as f:
            content = f.read()
            assert "www.bbottest.notreal,,,127.0.0.1" in content
        filename = next(module_test.scan.home.glob("asset-inventory-ip-addresses-table*.txt"))
        with open(filename) as f:
            assert "127.0.0.0/16" in f.read()
        filename = next(module_test.scan.home.glob("asset-inventory-domains-table*.txt"))
        with open(filename) as f:
            content = f.read()
            assert "bbottest.notreal" in content


class TestAsset_InventoryRecheck(TestAsset_Inventory):
    config_overrides = {
        "dns": {"minimal": False},
        "modules": {"asset_inventory": {"use_previous": True, "recheck": True}},
    }
    modules_overrides = ["asset_inventory"]

    def check(self, module_test, events):
        assert not any(e.type == "OPEN_TCP_PORT" for e in events), "Open port was emitted"
        assert any(e.data == "www.bbottest.notreal" for e in events), "No DNS name found"
        filename = next(module_test.scan.home.glob("asset-inventory.csv"))
        with open(filename) as f:
            content = f.read()
            assert "www.bbottest.notreal,,,127.0.0.1" in content



import ssl
import anyio
import httpx
import asyncio
import logging
import traceback
from socksio.exceptions import SOCKSError
from contextlib import asynccontextmanager

from bbot.core.engine import EngineServer
from bbot.core.helpers.misc import bytes_to_human, human_to_bytes, get_exception_chain

log = logging.getLogger("bbot.core.helpers.web.engine")


class HTTPEngine(EngineServer):

    CMDS = {
        0: "request",
        1: "request_batch",
        2: "request_custom_batch",
        3: "download",
    }

    client_only_options = (
        "retries",
        "max_redirects",
    )

    def __init__(self, socket_path, target, config={}, debug=False):
        super().__init__(socket_path, debug=debug)
        self.target = target
        self.config = config
        self.web_config = self.config.get("web", {})
        self.http_debug = self.web_config.get("debug", False)
        self._ssl_context_noverify = None
        self.web_clients = {}
        self.web_clients[0] = self.AsyncClient(persist_cookies=False, retries=0)
        self.web_client = self.web_clients[0]

    def AsyncClient(self, *args, **kwargs):
        # cache by retries to prevent unwanted accumulation of clients
        # (they are not garbage-collected)
        retries = kwargs.get("retries", 0)
        try:
            return self.web_clients[retries]
        except KeyError:
            from .client import BBOTAsyncClient

            client = BBOTAsyncClient.from_config(self.config, self.target, *args, **kwargs)
            self.web_clients[retries] = client
            return client

    async def request(self, *args, **kwargs):
        raise_error = kwargs.pop("raise_error", False)
        # TODO: use this
        cache_for = kwargs.pop("cache_for", None)  # noqa

        client = kwargs.get("client", self.web_client)

        # allow vs follow, httpx why??
        allow_redirects = kwargs.pop("allow_redirects", None)
        if allow_redirects is not None and "follow_redirects" not in kwargs:
            kwargs["follow_redirects"] = allow_redirects

        # in case of URL only, assume GET request
        if len(args) == 1:
            kwargs["url"] = args[0]
            args = []

        url = kwargs.get("url", "")

        if not args and "method" not in kwargs:
            kwargs["method"] = "GET"

        client_kwargs = {}
        for k in list(kwargs):
            if k in self.client_only_options:
                v = kwargs.pop(k)
                client_kwargs[k] = v

        if client_kwargs:
            client = self.AsyncClient(**client_kwargs)

        async with self._acatch(url, raise_error):
            if self.http_debug:
                logstr = f"Web request: {str(args)}, {str(kwargs)}"
                log.trace(logstr)
            response = await client.request(*args, **kwargs)
            if self.http_debug:
                log.trace(
                    f"Web response from {url}: {response} (Length: {len(response.content)}) headers: {response.headers}"
                )
            return response

    async def request_batch(self, urls, *args, threads=10, **kwargs):
        tasks = {}
        client_id = self.client_id_var.get()

        urls = list(urls)

        def new_task():
            if urls:
                url = urls.pop(0)
                task = self.new_child_task(client_id, self.request(url, *args, **kwargs))
                tasks[task] = url

        for _ in range(threads):  # Start initial batch of tasks
            new_task()

        while tasks:  # While there are tasks pending
            # Wait for the first task to complete
            finished = await self.finished_tasks(client_id, timeout=120)

            for task in finished:
                response = task.result()
                url = tasks.pop(task)
                yield (url, response)
                new_task()

    async def request_custom_batch(self, urls_and_kwargs, threads=10):
        tasks = {}
        client_id = self.client_id_var.get()
        urls_and_kwargs = list(urls_and_kwargs)

        def new_task():
            if urls_and_kwargs:  # Ensure there are args to process
                url, kwargs, custom_tracker = urls_and_kwargs.pop(0)
                task = self.new_child_task(client_id, self.request(url, **kwargs))
                tasks[task] = (url, kwargs, custom_tracker)

        for _ in range(threads):  # Start initial batch of tasks
            new_task()

        while tasks:  # While there are tasks pending
            # Wait for the first task to complete
            done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

            for task in done:
                response = task.result()
                url, kwargs, custom_tracker = tasks.pop(task)
                yield (url, kwargs, custom_tracker, response)
                new_task()

    async def download(self, url, **kwargs):
        warn = kwargs.pop("warn", True)
        filename = kwargs.pop("filename")
        raise_error = kwargs.get("raise_error", False)
        try:
            result = await self.stream_request(url, **kwargs)
            if result is None:
                raise httpx.HTTPError(f"No response from {url}")
            content, response = result
            log.debug(f"Download result: HTTP {response.status_code}")
            response.raise_for_status()
            with open(filename, "wb") as f:
                f.write(content)
            return filename
        except httpx.HTTPError as e:
            log_fn = log.verbose
            if warn:
                log_fn = log.warning
            log_fn(f"Failed to download {url}: {e}")
            if raise_error:
                raise

    async def stream_request(self, url, **kwargs):
        follow_redirects = kwargs.pop("follow_redirects", True)
        max_size = kwargs.pop("max_size", None)
        raise_error = kwargs.pop("raise_error", False)
        if max_size is not None:
            max_size = human_to_bytes(max_size)
        kwargs["follow_redirects"] = follow_redirects
        if not "method" in kwargs:
            kwargs["method"] = "GET"
        try:
            total_size = 0
            chunk_size = 8192
            chunks = []

            async with self._acatch(url, raise_error=True), self.web_client.stream(url=url, **kwargs) as response:
                agen = response.aiter_bytes(chunk_size=chunk_size)
                async for chunk in agen:
                    _chunk_size = len(chunk)
                    if max_size is not None and total_size + _chunk_size > max_size:
                        log.verbose(
                            f"Size of response from {url} exceeds {bytes_to_human(max_size)}, file will be truncated"
                        )
                        agen.aclose()
                        break
                    total_size += _chunk_size
                    chunks.append(chunk)
                return b"".join(chunks), response
        except httpx.HTTPError as e:
            self.log.debug(f"Error requesting {url}: {e}")
            if raise_error:
                raise

    def ssl_context_noverify(self):
        if self._ssl_context_noverify is None:
            ssl_context = ssl.create_default_context()
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE
            ssl_context.options &= ~ssl.OP_NO_SSLv2 & ~ssl.OP_NO_SSLv3
            ssl_context.set_ciphers("ALL:@SECLEVEL=0")
            ssl_context.options |= 0x4  # Add the OP_LEGACY_SERVER_CONNECT option
            self._ssl_context_noverify = ssl_context
        return self._ssl_context_noverify

    @asynccontextmanager
    async def _acatch(self, url, raise_error):
        """
        Asynchronous context manager to handle various httpx errors during a request.

        Yields:
            None

        Note:
            This function is internal and should generally not be used directly.
            `url`, `args`, `kwargs`, and `raise_error` should be in the same context as this function.
        """
        try:
            yield
        except httpx.TimeoutException:
            if raise_error:
                raise
            else:
                log.verbose(f"HTTP timeout to URL: {url}")
        except httpx.ConnectError:
            if raise_error:
                raise
            else:
                log.debug(f"HTTP connect failed to URL: {url}")
        except httpx.HTTPError as e:
            if raise_error:
                raise
            else:
                log.trace(f"Error with request to URL: {url}: {e}")
                log.trace(traceback.format_exc())
        except ssl.SSLError as e:
            msg = f"SSL error with request to URL: {url}: {e}"
            if raise_error:
                raise httpx.RequestError(msg)
            else:
                log.trace(msg)
                log.trace(traceback.format_exc())
        except anyio.EndOfStream as e:
            msg = f"AnyIO error with request to URL: {url}: {e}"
            if raise_error:
                raise httpx.RequestError(msg)
            else:
                log.trace(msg)
                log.trace(traceback.format_exc())
        except SOCKSError as e:
            msg = f"SOCKS error with request to URL: {url}: {e}"
            if raise_error:
                raise httpx.RequestError(msg)
            else:
                log.trace(msg)
                log.trace(traceback.format_exc())
        except BaseException as e:
            # don't log if the error is the result of an intentional cancellation
            if not any(isinstance(_e, asyncio.exceptions.CancelledError) for _e in get_exception_chain(e)):
                log.trace(f"Unhandled exception with request to URL: {url}: {e}")
                log.trace(traceback.format_exc())
            raise



from baddns.base import get_all_modules
from baddns.lib.loader import load_signatures
from .base import BaseModule

import asyncio
import logging


class baddns(BaseModule):
    watched_events = ["DNS_NAME", "DNS_NAME_UNRESOLVED"]
    produced_events = ["FINDING", "VULNERABILITY"]
    flags = ["active", "safe", "web-basic", "baddns", "cloud-enum", "subdomain-hijack"]
    meta = {
        "description": "Check hosts for domain/subdomain takeovers",
        "created_date": "2024-01-18",
        "author": "@liquidsec",
    }
    options = {"custom_nameservers": [], "only_high_confidence": False, "enable_references": False}
    options_desc = {
        "custom_nameservers": "Force BadDNS to use a list of custom nameservers",
        "only_high_confidence": "Do not emit low-confidence or generic detections",
        "enable_references": "Enable the references module (off by default)",
    }
    module_threads = 8
    deps_pip = ["baddns~=1.1.798"]

    def select_modules(self):

        module_list = ["CNAME", "NS", "MX", "TXT"]
        if self.config.get("enable_references", False):
            module_list.append("references")

        selected_modules = []
        for m in get_all_modules():
            if m.name in module_list:
                selected_modules.append(m)
        return selected_modules

    async def setup(self):
        self.preset.core.logger.include_logger(logging.getLogger("baddns"))
        self.custom_nameservers = self.config.get("custom_nameservers", []) or None
        if self.custom_nameservers:
            self.custom_nameservers = self.helpers.chain_lists(self.custom_nameservers)
        self.only_high_confidence = self.config.get("only_high_confidence", False)
        self.signatures = load_signatures()
        return True

    async def handle_event(self, event):

        tasks = []
        for ModuleClass in self.select_modules():
            module_instance = ModuleClass(
                event.data,
                http_client_class=self.scan.helpers.web.AsyncClient,
                dns_client=self.scan.helpers.dns.resolver,
                custom_nameservers=self.custom_nameservers,
                signatures=self.signatures,
            )
            tasks.append((module_instance, asyncio.create_task(module_instance.dispatch())))

        for module_instance, task in tasks:
            if await task:
                results = module_instance.analyze()
                if results and len(results) > 0:
                    for r in results:
                        r_dict = r.to_dict()

                        if r_dict["confidence"] in ["CONFIRMED", "PROBABLE"]:
                            data = {
                                "severity": "MEDIUM",
                                "description": f"{r_dict['description']}. Confidence: [{r_dict['confidence']}] Signature: [{r_dict['signature']}] Indicator: [{r_dict['indicator']}] Trigger: [{r_dict['trigger']}] baddns Module: [{r_dict['module']}]",
                                "host": str(event.host),
                            }
                            await self.emit_event(
                                data,
                                "VULNERABILITY",
                                event,
                                tags=[f"baddns-{module_instance.name.lower()}"],
                                context=f'{{module}}\'s "{r_dict["module"]}" module found {{event.type}}: {r_dict["description"]}',
                            )

                        elif r_dict["confidence"] in ["UNLIKELY", "POSSIBLE"] and not self.only_high_confidence:
                            data = {
                                "description": f"{r_dict['description']} Confidence: [{r_dict['confidence']}] Signature: [{r_dict['signature']}] Indicator: [{r_dict['indicator']}] Trigger: [{r_dict['trigger']}] baddns Module: [{r_dict['module']}]",
                                "host": str(event.host),
                            }
                            await self.emit_event(
                                data,
                                "FINDING",
                                event,
                                tags=[f"baddns-{module_instance.name.lower()}"],
                                context=f'{{module}}\'s "{r_dict["module"]}" module found {{event.type}}: {r_dict["description"]}',
                            )
                        else:
                            self.warning(f"Got unrecognized confidence level: {r['confidence']}")

                        found_domains = r_dict.get("found_domains", None)
                        if found_domains:
                            for found_domain in found_domains:
                                await self.emit_event(
                                    found_domain,
                                    "DNS_NAME",
                                    event,
                                    tags=[f"baddns-{module_instance.name.lower()}"],
                                    context=f'{{module}}\'s "{r_dict["module"]}" module found {{event.type}}: {{event.data}}',
                                )



import os
from pathlib import Path

from bbot.modules.base import BaseModule


class unstructured(BaseModule):
    watched_events = ["FILESYSTEM"]
    produced_events = ["FILESYSTEM", "RAW_TEXT"]
    flags = ["passive", "safe"]
    meta = {
        "description": "Module to extract data from files",
        "created_date": "2024-06-03",
        "author": "@domwhewell-sage",
    }
    options = {
        "extensions": [
            "bak",  #  Backup File
            "bash",  #  Bash Script or Configuration
            "bashrc",  #  Bash Script or Configuration
            "conf",  #  Configuration File
            "cfg",  #  Configuration File
            "crt",  #  Certificate File
            "csv",  #  Comma Separated Values File
            "db",  #  SQLite Database File
            "sqlite",  #  SQLite Database File
            "doc",  #  Microsoft Word Document (Old Format)
            "docx",  #  Microsoft Word Document
            "ica",  #  Citrix Independent Computing Architecture File
            "indd",  #  Adobe InDesign Document
            "ini",  #  Initialization File
            "key",  #  Private Key File
            "pub",  #  Public Key File
            "log",  #  Log File
            "markdown",  #  Markdown File
            "md",  #  Markdown File
            "odg",  #  OpenDocument Graphics (LibreOffice, OpenOffice)
            "odp",  #  OpenDocument Presentation (LibreOffice, OpenOffice)
            "ods",  #  OpenDocument Spreadsheet (LibreOffice, OpenOffice)
            "odt",  #  OpenDocument Text (LibreOffice, OpenOffice)
            "pdf",  #  Adobe Portable Document Format
            "pem",  #  Privacy Enhanced Mail (SSL certificate)
            "pps",  #  Microsoft PowerPoint Slideshow (Old Format)
            "ppsx",  #  Microsoft PowerPoint Slideshow
            "ppt",  #  Microsoft PowerPoint Presentation (Old Format)
            "pptx",  #  Microsoft PowerPoint Presentation
            "ps1",  #  PowerShell Script
            "rdp",  #  Remote Desktop Protocol File
            "sh",  #  Shell Script
            "sql",  #  SQL Database Dump
            "swp",  #  Swap File (temporary file, often Vim)
            "sxw",  #  OpenOffice.org Writer document
            "txt",  #  Plain Text Document
            "vbs",  #  Visual Basic Script
            "wpd",  #  WordPerfect Document
            "xls",  #  Microsoft Excel Spreadsheet (Old Format)
            "xlsx",  #  Microsoft Excel Spreadsheet
            "xml",  #  eXtensible Markup Language File
            "yml",  #  YAML Ain't Markup Language
            "yaml",  #  YAML Ain't Markup Language
        ],
        "ignore_folders": [".git"],
    }
    options_desc = {
        "extensions": "File extensions to parse",
        "ignore_folders": "Subfolders to ignore when crawling downloaded folders",
    }

    deps_apt = ["libmagic-dev", "poppler-utils", "tesseract-ocr", "libreoffice", "pandoc"]
    deps_pip = ["unstructured[all-docs]>=0.5.15,<1.0"]

    scope_distance_modifier = 1

    async def setup(self):
        self.extensions = list(set([e.lower().strip(".") for e in self.config.get("extensions", [])]))
        self.ignored_folders = self.config.get("ignore_folders", [])
        # Do not send user statistics to the unstructured library
        os.environ["SCARF_NO_ANALYTICS"] = "true"
        return True

    async def filter_event(self, event):
        if "file" not in event.tags and "folder" not in event.tags:
            return False, "Event is not a file or folder"
        if "file" in event.tags:
            if not any(event.data["path"].endswith(f".{ext}") for ext in self.extensions):
                return False, "File extension not in the allowed list"
        return True

    async def handle_event(self, event):
        if "folder" in event.tags:
            folder_path = Path(event.data["path"])
            for file_path in folder_path.rglob("*"):
                # If the file is not in an ignored folder and if it has an allowed extension raise it as a FILESYSTEM event
                if not any(ignored_folder in str(file_path) for ignored_folder in self.ignored_folders):
                    if any(file_path.name.endswith(f".{ext}") for ext in self.extensions):
                        file_event = self.make_event(
                            {"path": str(file_path)}, "FILESYSTEM", tags=["parsed_folder", "file"], parent=event
                        )
                        await self.emit_event(file_event)
        elif "file" in event.tags:
            file_path = event.data["path"]
            content = await self.scan.helpers.run_in_executor_mp(extract_text, file_path)
            if content:
                raw_text_event = self.make_event(
                    content,
                    "RAW_TEXT",
                    context=f"Extracted text from {file_path}",
                    parent=event,
                )
                await self.emit_event(raw_text_event)

    async def finish(self):
        del os.environ["SCARF_NO_ANALYTICS"]
        return


def extract_text(file_path):
    """
    extract_text Extracts plaintext from a document path using unstructured.

    :param file_path: The path of the file to extract text from.
    :return: ASCII-encoded plaintext extracted from the document.
    """

    from unstructured.partition.auto import partition

    unstructured_file_types = [
        ".csv",
        ".eml",
        ".msg",
        ".epub",
        ".xlsx",
        ".xls",
        ".html",
        ".htm",
        ".md",
        ".org",
        ".odt",
        ".pdf",
        ".txt",
        ".text",
        ".log",
        ".ppt",
        ".pptx",
        ".rst",
        ".rtf",
        ".tsv",
        ".doc",
        ".docx",
        ".xml",
    ]

    # If the file can be extracted with unstructured use its partition function or try and read it
    if any(file_path.lower().endswith(file_type) for file_type in unstructured_file_types):
        try:
            elements = partition(filename=file_path)
            return "\n\n".join(element.text for element in elements)
        except ValueError:
            with open(file_path, "rb") as file:
                return file.read().decode("utf-8", errors="ignore")
    else:
        with open(file_path, "rb") as file:
            return file.read().decode("utf-8", errors="ignore")



import json
from contextlib import suppress
from neo4j import AsyncGraphDatabase

from bbot.modules.output.base import BaseOutputModule


class neo4j(BaseOutputModule):
    """
    # start Neo4j in the background with docker
    docker run -d -p 7687:7687 -p 7474:7474 -v "$(pwd)/neo4j/:/data/" -e NEO4J_AUTH=neo4j/bbotislife neo4j

    # view all running docker containers
    > docker ps

    # view all docker containers
    > docker ps -a

    # stop a docker container
    > docker stop <CONTAINER_ID>

    # remove a docker container
    > docker remove <CONTAINER_ID>

    # start a stopped container
    > docker start <CONTAINER_ID>
    """

    watched_events = ["*"]
    meta = {"description": "Output to Neo4j", "created_date": "2022-04-07", "author": "@TheTechromancer"}
    options = {"uri": "bolt://localhost:7687", "username": "neo4j", "password": "bbotislife"}
    options_desc = {
        "uri": "Neo4j server + port",
        "username": "Neo4j username",
        "password": "Neo4j password",
    }
    deps_pip = ["neo4j"]
    _batch_size = 500
    _preserve_graph = True

    async def setup(self):
        try:
            self.driver = AsyncGraphDatabase.driver(
                uri=self.config.get("uri", self.options["uri"]),
                auth=(
                    self.config.get("username", self.options["username"]),
                    self.config.get("password", self.options["password"]),
                ),
            )
            self.session = self.driver.session()
            await self.handle_event(self.scan.root_event)
        except Exception as e:
            return False, f"Error setting up Neo4j: {e}"
        return True

    async def handle_batch(self, *all_events):
        # group events by type, since cypher doesn't allow dynamic labels
        events_by_type = {}
        parents_by_type = {}
        relationships = []
        for event in all_events:
            parent = event.get_parent()
            try:
                events_by_type[event.type].append(event)
            except KeyError:
                events_by_type[event.type] = [event]
            try:
                parents_by_type[parent.type].append(parent)
            except KeyError:
                parents_by_type[parent.type] = [parent]

            module = str(event.module)
            timestamp = event.timestamp
            relationships.append((parent, module, timestamp, event))

        all_ids = {}
        for event_type, events in events_by_type.items():
            self.debug(f"{len(events):,} events of type {event_type}")
            all_ids.update(await self.merge_events(events, event_type))
        for event_type, parents in parents_by_type.items():
            self.debug(f"{len(parents):,} parents of type {event_type}")
            all_ids.update(await self.merge_events(parents, event_type, id_only=True))

        rel_ids = []
        for parent, module, timestamp, event in relationships:
            try:
                src_id = all_ids[parent.id]
                dst_id = all_ids[event.id]
            except KeyError as e:
                self.error(f'Error "{e}" correlating {parent.id}:{parent.data} --> {event.id}:{event.data}')
                continue
            rel_ids.append((src_id, module, timestamp, dst_id))

        await self.merge_relationships(rel_ids)

    async def merge_events(self, events, event_type, id_only=False):
        if id_only:
            insert_data = [{"data": str(e.data), "type": e.type, "id": e.id} for e in events]
        else:
            insert_data = []
            for e in events:
                event_json = e.json(mode="graph")
                # we pop the timestamp because it belongs on the relationship
                event_json.pop("timestamp")
                # nested data types aren't supported in neo4j
                for key in ("dns_children", "discovery_path"):
                    if key in event_json:
                        event_json[key] = json.dumps(event_json[key])
                insert_data.append(event_json)

        cypher = f"""UNWIND $events AS event
        MERGE (_:{event_type} {{ id: event.id }})
        SET _ += event
        RETURN event.data as event_data, event.id as event_id, elementId(_) as neo4j_id"""
        neo4j_ids = {}
        # insert events
        try:
            results = await self.session.run(cypher, events=insert_data)
            # get Neo4j ids
            for result in await results.data():
                event_id = result["event_id"]
                neo4j_id = result["neo4j_id"]
                neo4j_ids[event_id] = neo4j_id
        except Exception as e:
            self.error(f"Error inserting Neo4j nodes (label:{event_type}): {e}")
            self.trace(insert_data)
            self.trace(cypher)
        return neo4j_ids

    async def merge_relationships(self, relationships):
        rels_by_module = {}
        # group by module
        for src_id, module, timestamp, dst_id in relationships:
            data = {"src_id": src_id, "timestamp": timestamp, "dst_id": dst_id}
            try:
                rels_by_module[module].append(data)
            except KeyError:
                rels_by_module[module] = [data]

        for module, rels in rels_by_module.items():
            self.debug(f"{len(rels):,} relationships of type {module}")
            cypher = f"""
            UNWIND $rels AS rel
            MATCH (a) WHERE elementId(a) = rel.src_id
            MATCH (b) WHERE elementId(b) = rel.dst_id
            MERGE (a)-[_:{module}]->(b)
            SET _.timestamp = rel.timestamp"""
            try:
                await self.session.run(cypher, rels=rels)
            except Exception as e:
                self.error(f"Error inserting Neo4j relationship (label:{module}): {e}")
                self.trace(cypher)

    async def cleanup(self):
        with suppress(Exception):
            await self.session.close()
        with suppress(Exception):
            await self.driver.close()
